<?php

namespace Drupal\bpmn_io\Plugin\modeler_api_modeler;

use Drupal\bpmn_io\Service\Parser;
use Drupal\bpmn_io\Service\PrepareComponents;
use Drupal\Component\Utility\Random;
use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Form\FormBuilder;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\bpmn_io\Form\Modeler as ModelerForm;
use Drupal\bpmn_io\Service\TokenBrowserService;
use Drupal\modeler_api\Api;
use Drupal\modeler_api\Attribute\Modeler;
use Drupal\modeler_api\Plugin\modeler_api_model_owner\ModelOwnerInterface;
use Drupal\modeler_api\Plugin\modeler_api_modeler\ModelerBase;
use Drupal\modeler_api\Plugin\modeler_api_modeler\ModelerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the ECA Modeller.
 */
#[Modeler(
  id: "bpmn_io",
  label: new TranslatableMarkup("BPMN.iO"),
  description: new TranslatableMarkup("BPMN modeller with a feature-rich UI.")
)]
class BpmnIo extends ModelerBase {

  public const array SUPPORTED_COMPONENT_TYPES = [
    Api::COMPONENT_TYPE_START => 'bpmn:Event',
    Api::COMPONENT_TYPE_CONDITION => 'bpmn:SequenceFlow',
    Api::COMPONENT_TYPE_ELEMENT => 'bpmn:Task',
    Api::COMPONENT_TYPE_LINK => 'bpmn:Gateway',
    // Api::COMPONENT_TYPE_GROUP => '', To be added later.
  ];

  public const array MAP_COMPONENT_TYPES = [
    'bpmn:Event' => 'StartEvent',
    'bpmn:SequenceFlow' => 'SequenceFlow',
    'bpmn:Task' => 'Task',
    'bpmn:Gateway' => 'ExclusiveGateway',
    // 'group', To be added later.
  ];

  /**
   * Dependency Injection container.
   *
   * Used for getter injection.
   *
   * @var \Symfony\Component\DependencyInjection\ContainerInterface|null
   */
  protected ?ContainerInterface $container;

  /**
   * The form builder.
   *
   * @var \Drupal\Core\Form\FormBuilder
   */
  protected FormBuilder $formBuilder;

  /**
   * The token browser service.
   *
   * @var \Drupal\bpmn_io\Service\TokenBrowserService
   */
  protected TokenBrowserService $tokenBrowserService;

  /**
   * The modeler API.
   *
   * @var \Drupal\modeler_api\Api
   */
  protected Api $modelerApi;

  /**
   * The prepare components service.
   *
   * @var \Drupal\bpmn_io\Service\PrepareComponents
   */
  protected PrepareComponents $prepareComponents;

  /**
   * The BPMN parser.
   *
   * @var \Drupal\bpmn_io\Service\Parser
   */
  protected Parser $parser;

  /**
   * Get Dependency Injection container.
   *
   * @return \Symfony\Component\DependencyInjection\ContainerInterface
   *   Current Dependency Injection container.
   */
  protected function getContainer(): ContainerInterface {
    if (!isset($this->container)) {
      // @phpstan-ignore-next-line
      $this->container = \Drupal::getContainer();
    }
    return $this->container;
  }

  /**
   * Get the form builder.
   *
   * @return \Drupal\Core\Form\FormBuilder
   *   The form builder.
   */
  protected function formBuilder(): FormBuilder {
    if (!isset($this->formBuilder)) {
      $this->formBuilder = $this->getContainer()->get('form_builder');
    }
    return $this->formBuilder;
  }

  /**
   * Get the token browser service.
   *
   * @return \Drupal\bpmn_io\Service\TokenBrowserService
   *   The token browser service.
   */
  protected function tokenBrowserService(): TokenBrowserService {
    if (!isset($this->tokenBrowserService)) {
      $this->tokenBrowserService = $this->getContainer()->get('bpmn_io.token_browser');
    }
    return $this->tokenBrowserService;
  }

  /**
   * Get the modeler API.
   *
   * @return \Drupal\modeler_api\Api
   *   The modeler API.
   */
  protected function modelerApi(): Api {
    if (!isset($this->modelerApi)) {
      $this->modelerApi = $this->getContainer()->get('modeler_api.service');
    }
    return $this->modelerApi;
  }

  /**
   * Get the prepare components service.
   *
   * @return \Drupal\bpmn_io\Service\PrepareComponents
   *   The prepare components service.
   */
  protected function prepareComponents(): PrepareComponents {
    if (!isset($this->prepareComponents)) {
      $this->prepareComponents = $this->getContainer()->get('bpmn_io.prepare_components');
    }
    return $this->prepareComponents;
  }

  /**
   * Get the BPMN parser.
   *
   * @return \Drupal\bpmn_io\Service\Parser
   *   The BPMN parser.
   */
  protected function parser(): Parser {
    if (!isset($this->parser)) {
      $this->parser = $this->getContainer()->get('bpmn_io.parser');
    }
    return $this->parser;
  }

  /**
   * {@inheritdoc}
   */
  public function getRawFileExtension(): ?string {
    return 'xml';
  }

  /**
   * {@inheritdoc}
   */
  public function isEditable(): bool {
    return TRUE;
  }

  /**
   * Returns a render array with everything required for model editing.
   *
   * @return array
   *   The render array.
   */
  public function edit(ModelOwnerInterface $owner, string $id, string $data, bool $isNew = FALSE): array {
    $form = $this->formBuilder()->getForm(ModelerForm::class, $owner, $id);
    if (isset($form['gin_sidebar'])) {
      $form['gin_sidebar']['property_panel'] = ['#markup' => '<div class="property-panel"></div>'];
      $form['gin_sidebar']['token_browser'] = $this->tokenBrowserService()->getTokenBrowserMarkup();
      $extras = '';
    }
    else {
      $extras = '<div class="property-panel in-canvas"></div>';
      $form['actions']['tb'] = $this->tokenBrowserService()->getTokenBrowserMarkup();
    }
    return [
      '#type' => 'container',
      '#attributes' => [
        'id' => 'bpmn-io',
      ],
      'canvas' => [
        '#prefix' => '<div class="canvas" role="application" aria-label="BPMN Canvas"></div>' . $extras,
      ],
      'form' => $form,
      '#attached' => [
        'library' => [
          'bpmn_io/ui',
        ],
        'drupalSettings' => [
          'bpmn_io' => [
            'id' => $id,
            'is_new' => $isNew,
            'modeller' => 'bpmn_io',
            'bpmn' => $data,
            'templates' => $this->prepareComponents()->getTemplates($owner),
            'save_url' => Url::fromRoute('entity.' . $owner->configEntityTypeId() . '.save', ['modeler_id' => 'bpmn_io'])->toString(),
            'token_url' => Url::fromRoute('system.csrftoken')->toString(),
            'collection_url' => Url::fromRoute('entity.' . $owner->configEntityTypeId() . '.collection')->toString(),
          ],
        ],
      ],
    ];
  }

  /**
   * Returns a render array with everything required for model editing.
   *
   * @return array
   *   The render array.
   */
  public function convert(ModelOwnerInterface $owner, ConfigEntityInterface $model): array {
    $reload = FALSE;
    if (!in_array($owner->getModelerId($model), ['bpmn_io', 'fallback'])) {
      // Let the modeller clone the ECA-entity.
      $model = $owner->clone($model);
      $reload = TRUE;
    }
    $owner->setModelerId($model, 'bpmn_io');
    $model->save();
    $type = $model->getEntityTypeId();

    $id = '';
    $emptyModelData = $this->prepareEmptyModelData($id);
    $emptyModelData = str_replace($id, $model->id(), $emptyModelData);

    // Provide enough mappings so that the Javascript-class can do its thing.
    $build = $this->edit($owner, $model->id(), $emptyModelData);
    $build['canvas']['#suffix'] = '<div class="convert-overlay">' . $this->t('Conversion in progress ...') . '</div>';
    $build['#attached']['library'] = 'bpmn_io/convert';
    $build['#attached']['drupalSettings']['bpmn_io_convert'] = [
      'reload' => $reload,
      'metadata' => [
        'name' => $model->label(),
        'version' => $model->get('version'),
        'redirect_url' => Url::fromRoute(
          'entity.' . $type . '.edit_form',
          [$type => $model->id()],
          ['absolute' => TRUE]
        )->toString(),
      ],
    ];

    $components = $owner->usedComponents($model);
    $supportedOwnerComponentTypes = $owner->supportedOwnerComponentTypes();
    $elements = $bpmnMapping = $templateMapping = [];
    foreach ($components as $component) {
      if (!isset(self::SUPPORTED_COMPONENT_TYPES[$component->getType()])) {
        // @todo Log this missing component type.
        continue;
      }
      $successors = [];
      foreach ($component->getSuccessors() as $successor) {
        $successors[] = [
          'id' => $successor->getId(),
          'condition' => $successor->getConditionId(),
        ];
      }
      $elements[$component->getId()] = [
        'plugin' => $component->getPluginId(),
        'label' => $component->getLabel(),
        'configuration' => $component->getConfiguration(),
        'successors' => $successors,
      ];
      $bpmnMapping[$component->getId()] = self::MAP_COMPONENT_TYPES[self::SUPPORTED_COMPONENT_TYPES[$component->getType()]];
      $templateMapping[$component->getId()] = $supportedOwnerComponentTypes[$component->getType()];
    }
    $build['#attached']['drupalSettings']['bpmn_io_convert']['elements'] = $elements;
    $build['#attached']['drupalSettings']['bpmn_io_convert']['bpmn_mapping'] = $bpmnMapping;
    $build['#attached']['drupalSettings']['bpmn_io_convert']['template_mapping'] = $templateMapping;

    return $build;
  }

  /**
   * {@inheritdoc}
   */
  public function generateId(): string {
    $random = new Random();
    return 'Process_' . $random->name(7);
  }

  /**
   * {@inheritdoc}
   */
  public function enable(): ModelerInterface {
    $this->parser()->enable();
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function disable(): ModelerInterface {
    $this->parser()->disable();
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function clone(string $id, string $label): ModelerInterface {
    $this->parser()->clone($id, $label);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function prepareEmptyModelData(string &$id): string {
    $id = $this->generateId();
    $emptyBpmn = file_get_contents($this->extensionPathResolver->getPath('module', 'bpmn_io') . DIRECTORY_SEPARATOR . 'data' . DIRECTORY_SEPARATOR . 'empty.bpmn');
    return str_replace([
      'SIDPLACEHOLDER1',
      'SIDPLACEHOLDER2',
      'IDPLACEHOLDER',
    ], [
      'sid-' . $this->uuid->generate(),
      'sid-' . $this->uuid->generate(),
      $id,
    ], $emptyBpmn);
  }

  /**
   * {@inheritdoc}
   */
  public function getId(): string {
    return $this->parser()->getId();
  }

  /**
   * {@inheritdoc}
   */
  public function getLabel(): string {
    return $this->parser()->getLabel();
  }

  /**
   * {@inheritdoc}
   */
  public function getTags(): array {
    return $this->parser()->getTags();
  }

  /**
   * {@inheritdoc}
   */
  public function getChangelog(): array {
    return $this->parser()->getChangelog();
  }

  /**
   * {@inheritdoc}
   */
  public function getDocumentation(): string {
    return $this->parser()->getDocumentation();
  }

  /**
   * {@inheritdoc}
   */
  public function getStatus(): bool {
    return $this->parser()->getStatus();
  }

  /**
   * {@inheritdoc}
   */
  public function getVersion(): string {
    return $this->parser()->getVersion();
  }

  /**
   * {@inheritdoc}
   */
  public function parseData(string $data): void {
    parent::parseData($data);
    $this->parser()->setData($data);
  }

  /**
   * {@inheritdoc}
   */
  public function readComponents(): array {
    return $this->parser()->getComponents();
  }

  /**
   * {@inheritdoc}
   */
  public function updateComponents(ModelOwnerInterface $owner): bool {
    $changed = $this->parser()->updateComponents($this->prepareComponents()->getTemplates($owner));
    $this->parseData($this->parser()->getData());
    return $changed;
  }

}
