<?php

namespace Drupal\bpmn_io\Service;

use Drupal\modeler_api\Api;
use Drupal\modeler_api\Component;
use Drupal\modeler_api\ComponentSuccessor;
use Mtownsend\XmlToArray\XmlToArray;

/**
 * The parser service for BPMN XML data.
 */
class Parser {

  public const int GATEWAY_TYPE_EXCLUSIVE = 0;
  public const int GATEWAY_TYPE_PARALLEL = 1;
  public const int GATEWAY_TYPE_INCLUSIVE = 2;
  public const int GATEWAY_TYPE_COMPLEX = 3;
  public const int GATEWAY_TYPE_EVENT_BASED = 4;

  /**
   * The raw XML string data.
   *
   * @var string
   */
  protected string $rawData;

  /**
   * The XML model.
   *
   * @var array
   */
  protected array $xmlModel;

  /**
   * The XML DOM.
   *
   * @var \DOMDocument
   */
  protected \DOMDocument $doc;

  /**
   * The XML DOM XPath.
   *
   * @var \DOMXPath
   */
  protected \DOMXPath $xpath;

  /**
   * The list of components.
   *
   * @var \Drupal\modeler_api\Component[]
   */
  protected array $components;

  /**
   * Get all components.
   *
   * @return \Drupal\modeler_api\Component[]
   *   All components.
   */
  public function getComponents(): array {
    return $this->components;
  }

  /**
   * Gets the ID of the model.
   *
   * @return string
   *   The ID.
   */
  public function getId(): string {
    return $this->xmlModel[$this->xmlNsPrefix() . 'process']['@attributes']['id'];
  }

  /**
   * Gets the label of the model.
   *
   * @return string
   *   The label.
   */
  public function getLabel(): string {
    return $this->xmlModel[$this->xmlNsPrefix() . 'process']['@attributes']['name'] ?? 'noname';
  }

  /**
   * Gets the tags of the model.
   *
   * @return array
   *   The tags.
   */
  public function getTags(): array {
    $process = $this->xmlNsPrefix() . 'process';
    $extensions = $this->xmlNsPrefix() . 'extensionElements';
    $tags = isset($this->xmlModel[$process][$extensions]) ?
      explode(',', $this->findProperty($this->xmlModel[$process][$extensions], 'Tags')) :
      [];
    array_walk($tags, static function (&$item) {
      $item = trim((string) $item);
    });
    return $tags;
  }

  /**
   * Gets the changelog of the model.
   *
   * @return array
   *   The changelog.
   */
  public function getChangelog(): array {
    $process = $this->xmlNsPrefix() . 'process';
    $extensions = $this->xmlNsPrefix() . 'extensionElements';
    $changelog = [];
    if (isset($this->xmlModel[$process][$extensions])) {
      $v = 1;
      while ($item = $this->findProperty($this->xmlModel[$process][$extensions], 'Changelog v' . $v)) {
        $changelog['v' . $v] = $item;
        $v++;
      }
    }
    return $changelog;
  }

  /**
   * Gets the documentation of the model.
   *
   * @return string
   *   The documentation.
   */
  public function getDocumentation(): string {
    return $this->xmlModel[$this->xmlNsPrefix() . 'process'][$this->xmlNsPrefix() . 'documentation'] ?? '';
  }

  /**
   * Gets the status of the model.
   *
   * @return bool
   *   The status.
   */
  public function getStatus(): bool {
    return mb_strtolower($this->xmlModel[$this->xmlNsPrefix() . 'process']['@attributes']['isExecutable'] ?? 'true') === 'true';
  }

  /**
   * Gets the version of the model.
   *
   * @return string
   *   The version.
   */
  public function getVersion(): string {
    return $this->xmlModel[$this->xmlNsPrefix() . 'process']['@attributes']['versionTag'] ?? '';
  }

  /**
   * Get the raw data.
   *
   * @return string
   *   The raw data.
   */
  public function getData(): string {
    return $this->rawData;
  }

  /**
   * Set and parse the data.
   *
   * @param string $data
   *   The raw data.
   */
  public function setData(string $data): void {
    $this->rawData = $data;
    $this->xmlModel = XmlToArray::convert($this->rawData);
    $this->doc = new \DOMDocument();
    $this->doc->loadXML($this->rawData);
    $this->xpath = new \DOMXPath($this->doc);

    $idxExtension = $this->xmlNsPrefix() . 'extensionElements';
    $this->components = [];

    $flow = [];
    foreach ($this->getSequenceFlows() as $sequenceFlow) {
      if (isset($sequenceFlow[$idxExtension])) {
        $pluginId = $this->findProperty($sequenceFlow[$idxExtension], 'pluginid');
        $condition = $this->findAttribute($sequenceFlow, 'id');
        if (!empty($pluginId) && !empty($condition)) {
          $this->components[] = new Component(
            $condition,
            Api::COMPONENT_TYPE_CONDITION,
            $pluginId,
            $this->findAttribute($sequenceFlow, 'name'),
            $this->findFields($sequenceFlow[$idxExtension]),
          );
        }
        else {
          $condition = '';
        }
      }
      else {
        $condition = '';
      }
      $flow[$this->findAttribute($sequenceFlow, 'sourceRef')][] = new ComponentSuccessor(
        $this->findAttribute($sequenceFlow, 'targetRef'),
        $condition,
      );
    }

    foreach ($this->getGateways() as $gateway) {
      $gatewayId = $this->findAttribute($gateway, 'id');
      $this->components[] = new Component(
        $gatewayId,
        Api::COMPONENT_TYPE_LINK,
        '',
        '',
        [],
        $flow[$gatewayId] ?? [],
      );
    }

    foreach ($this->getStartEvents() as $startEvent) {
      $extension = $startEvent[$idxExtension] ?? [];
      $pluginId = $this->findProperty($extension, 'pluginid');
      if (empty($pluginId)) {
        continue;
      }
      $id = $this->findAttribute($startEvent, 'id');
      $this->components[] = new Component(
        $id,
        Api::COMPONENT_TYPE_START,
        $pluginId,
        $this->findAttribute($startEvent, 'name'),
        $this->findFields($extension),
        $flow[$id] ?? []
      );
    }

    foreach ($this->getTasks() as $task) {
      $extension = $task[$idxExtension] ?? [];
      $pluginId = $this->findProperty($extension, 'pluginid');
      if (empty($pluginId)) {
        continue;
      }
      $id = $this->findAttribute($task, 'id');
      $this->components[] = new Component(
        $id,
        Api::COMPONENT_TYPE_ELEMENT,
        $pluginId,
        $this->findAttribute($task, 'name'),
        $this->findFields($extension),
        $flow[$id] ?? []
      );
    }
  }

  /**
   * Gets the bpmn XML namespace prefix.
   *
   * @return string
   *   The prefix.
   */
  private function xmlNsPrefix(): string {
    return 'bpmn2:';
  }

  /**
   * Returns all the sequenceFlow objects (condition) from the XML model.
   *
   * @return array
   *   The list of all sequence flows in the model data.
   */
  private function getSequenceFlows(): array {
    $conditions = $this->xmlModel[$this->xmlNsPrefix() . 'process'][$this->xmlNsPrefix() . 'sequenceFlow'] ?? [];
    if (isset($conditions['@attributes'])) {
      return [$conditions];
    }
    return $conditions;
  }

  /**
   * Returns all the startEvent (events) objects from the XML model.
   *
   * @return array
   *   The list of all start events in the model data.
   */
  private function getStartEvents(): array {
    $events = $this->xmlModel[$this->xmlNsPrefix() . 'process'][$this->xmlNsPrefix() . 'startEvent'] ?? [];
    if (isset($events['@attributes'])) {
      return [$events];
    }
    return $events;
  }

  /**
   * Returns all the task objects (actions) from the XML model.
   *
   * @return array
   *   The list of all tasks in the model data.
   */
  private function getTasks(): array {
    $actions = $this->xmlModel[$this->xmlNsPrefix() . 'process'][$this->xmlNsPrefix() . 'task'] ?? [];
    if (isset($actions['@attributes'])) {
      return [$actions];
    }
    return $actions;
  }

  /**
   * Returns all the gateway objects from the XML model.
   *
   * @return array
   *   The list of all gateways in the model data.
   */
  private function getGateways(): array {
    $types = [
      self::GATEWAY_TYPE_EXCLUSIVE => 'exclusiveGateway',
      self::GATEWAY_TYPE_PARALLEL => 'parallelGateway',
      self::GATEWAY_TYPE_INCLUSIVE => 'inclusiveGateway',
      self::GATEWAY_TYPE_COMPLEX => 'complexGateway',
      self::GATEWAY_TYPE_EVENT_BASED => 'eventBasedGateway',
    ];
    $gateways = [];
    foreach ($types as $key => $type) {
      $objects = $this->xmlModel[$this->xmlNsPrefix() . 'process'][$this->xmlNsPrefix() . $type] ?? [];
      if (isset($objects['@attributes'])) {
        $objects = [$objects];
      }
      foreach ($objects as $object) {
        $object['type'] = $key;
        $gateways[] = $object;
      }
    }
    return $gateways;
  }

  /**
   * Return a property of a given BPMN element.
   *
   * @param array $element
   *   The BPMN element from which the property should be returned.
   * @param string $property_name
   *   The name of the property in the BPMN element.
   *
   * @return string
   *   The property's value, default to an empty string.
   */
  private function findProperty(array $element, string $property_name): string {
    if (isset($element['camunda:properties']['camunda:property'])) {
      $elements = isset($element['camunda:properties']['camunda:property']['@attributes']) ?
        [$element['camunda:properties']['camunda:property']] :
        $element['camunda:properties']['camunda:property'];
      foreach ($elements as $child) {
        if ($child['@attributes']['name'] === $property_name) {
          return $child['@attributes']['value'];
        }
      }
    }
    return '';
  }

  /**
   * Return an attribute of a given BPMN element.
   *
   * @param array $element
   *   The BPMN element from which the attribute should be returned.
   * @param string $attribute_name
   *   The name of the attribute in the BPMN element.
   *
   * @return string
   *   The attribute's value, default to an empty string.
   */
  private function findAttribute(array $element, string $attribute_name): string {
    return $element['@attributes'][$attribute_name] ?? '';
  }

  /**
   * Return all the field values of a given BPMN element.
   *
   * @param array $element
   *   The BPMN element from which the field values should be returned.
   *
   * @return array
   *   An array containing all the field values, keyed by the field name.
   */
  private function findFields(array $element): array {
    $fields = [];
    if (isset($element['camunda:field'])) {
      $elements = isset($element['camunda:field']['@attributes']) ? [$element['camunda:field']] : $element['camunda:field'];
      foreach ($elements as $child) {
        $fields[$child['@attributes']['name']] = isset($child['camunda:string']) && is_string($child['camunda:string']) ? $child['camunda:string'] : '';
      }
    }
    return $fields;
  }

  /**
   * Updates all components in raw data.
   *
   * @param array $templates
   *   All available templates received from the model owner.
   *
   * @return bool
   *   TRUE, if the model got changed, FALSE otherwise.
   *
   * @see \Drupal\modeler_api\Plugin\ModelerInterface::updateComponents()
   */
  public function updateComponents(array $templates): bool {
    $changed = FALSE;
    $idxExtension = $this->xmlNsPrefix() . 'extensionElements';
    foreach ($templates as $template) {
      foreach ($template['appliesTo'] as $type) {
        $objects = match ($type) {
          'bpmn:Event' => $this->getStartEvents(),
          'bpmn:SequenceFlow' => $this->getSequenceFlows(),
          'bpmn:Task' => $this->getTasks(),
          default => [],
        };
        foreach ($objects as $object) {
          if (isset($object['@attributes']['modelerTemplate']) && $template['id'] === $object['@attributes']['modelerTemplate']) {
            $fields = $this->findFields($object[$idxExtension]);
            $id = $object['@attributes']['id'];
            /**
             * @var \DOMElement|null $element
             */
            $element = $this->xpath->query("//*[@id='$id']")->item(0);
            if ($element) {
              /**
               * @var \DOMElement|null $extensions
               */
              $extensions = $this->xpath->query("//*[@id='$id']/$idxExtension")
                ->item(0);
              if (!$extensions) {
                $node = $this->doc->createElement($idxExtension);
                $extensions = $element->appendChild($node);
              }
              foreach ($template['properties'] as $property) {
                switch ($property['binding']['type']) {
                  case 'camunda:property':
                    if ($this->findProperty($object[$idxExtension], $property['binding']['name']) !== $property['value']) {
                      $element->setAttribute($property['binding']['name'], $property['value']);
                      $changed = TRUE;
                    }
                    break;

                  case 'camunda:field':
                    if (isset($fields[$property['binding']['name']])) {
                      // Field exists, remove it from the list.
                      unset($fields[$property['binding']['name']]);
                    }
                    else {
                      $fieldNode = $this->doc->createElement('camunda:field');
                      $fieldNode->setAttribute('name', $property['binding']['name']);
                      $valueNode = $this->doc->createElement('camunda:string');
                      $valueNode->textContent = $property['value'];
                      $fieldNode->appendChild($valueNode);
                      $extensions->appendChild($fieldNode);
                      $changed = TRUE;
                    }
                    break;
                }
              }
              // Remove remaining fields from the model.
              foreach ($fields as $name => $value) {
                /**
                 * @var \DOMElement $fieldElement
                 */
                if ($fieldElement = $this->xpath->query("//*[@id='$id']/$idxExtension/camunda:field[@name='$name']")
                  ->item(0)) {
                  $extensions->removeChild($fieldElement);
                  $changed = TRUE;
                }
              }
            }
          }
        }
      }
    }
    if ($changed) {
      $this->setData($this->doc->saveXML());
    }
    return $changed;
  }

  /**
   * Set the executable attribute to true.
   */
  public function enable(): void {
    /** @var \DOMElement|null $element */
    $element = $this->xpath->query("//*[@id='{$this->getId()}']")->item(0);
    if ($element) {
      $element->setAttribute('isExecutable', 'true');
      $this->setData($this->doc->saveXML());
    }
  }

  /**
   * Set the executable attribute to false.
   */
  public function disable(): void {
    /** @var \DOMElement|null $element */
    $element = $this->xpath->query("//*[@id='{$this->getId()}']")->item(0);
    if ($element) {
      $element->setAttribute('isExecutable', 'false');
      $this->setData($this->doc->saveXML());
    }
  }

  /**
   * Prepare raw data for a cloned version of the model.
   *
   * @param string $id
   *   The new ID.
   * @param string $label
   *   The new label.
   */
  public function clone(string $id, string $label): void {
    /** @var \DOMElement|null $element */
    $element = $this->xpath->query("//*[@id='{$this->getId()}']")->item(0);
    if ($element) {
      $element->setAttribute('id', $id);
      $element->setAttribute('name', $label);
      $this->setData($this->doc->saveXML());
    }
  }

}
