<?php

namespace Drupal\bpmn_io\Service;

use Drupal\bpmn_io\Plugin\modeler_api_modeler\BpmnIo;
use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\eca\Plugin\ECA\EcaPluginBase;
use Drupal\modeler_api\Api;
use Drupal\modeler_api\Plugin\modeler_api_model_owner\ModelOwnerInterface;

/**
 * Provides services to prepare components for the bpmn_io modeler.
 */
readonly class PrepareComponents {

  /**
   * Constructs the prepare components service.
   */
  public function __construct(
    protected Api $modelerApi,
    protected LoggerChannelInterface $logger,
  ) {}

  /**
   * Returns all the templates for the modeller UI.
   *
   * This includes templates for events, conditions and actions.
   *
   * @param \Drupal\modeler_api\Plugin\modeler_api_model_owner\ModelOwnerInterface $owner
   *   The model owner.
   *
   * @return array
   *   The list of all templates.
   */
  public function getTemplates(ModelOwnerInterface $owner): array {
    $templates = [];
    foreach (BpmnIo::SUPPORTED_COMPONENT_TYPES as $componentType => $appliesTo) {
      foreach ($this->modelerApi->availableOwnerComponents($owner, $componentType) as $plugin) {
        $pluginType = $owner->ownerComponentId($componentType);
        $form = $owner->buildConfigurationForm($plugin);
        $docUrl = $owner->pluginDocUrl($plugin, $pluginType);
        $templates[] = $this->properties($plugin, $pluginType, $appliesTo, $form, $docUrl);
      }
    }
    return $templates;
  }

  /**
   * Helper function to build a template for an event, condition or action.
   *
   * @param \Drupal\Component\Plugin\PluginInspectionInterface $plugin
   *   The event, condition or action plugin for which the template should
   *   be build.
   * @param string $pluginType
   *   The string identifying the plugin type, which is one of event, condition
   *   or action.
   * @param string $appliesTo
   *   The string to tell the modeller, to which object type the template will
   *   apply. Valid values are "bpmn:Event", "bpmn:sequenceFlow" or "bpmn:task".
   * @param array $form
   *   An array containing the configuration form of the plugin.
   * @param string|null $docUrl
   *   The documentation URL if it exists, NULL otherwise.
   *
   * @return array
   *   The completed template for BPMN modellers for the given plugin and its
   *   fields.
   */
  protected function properties(PluginInspectionInterface $plugin, string $pluginType, string $appliesTo, array $form, ?string $docUrl): array {
    $properties = [
      [
        'label' => 'Plugin ID',
        'type' => 'Hidden',
        'value' => $plugin->getPluginId(),
        'binding' => [
          'type' => 'camunda:property',
          'name' => 'pluginid',
        ],
      ],
    ];
    $extraDescriptions = [];
    foreach ($this->prepareConfigFields($form, $extraDescriptions) as $field) {
      if (!isset($field['value'])) {
        $value = '';
      }
      elseif (is_scalar($field['value'])) {
        $value = (string) $field['value'];
      }
      elseif (is_object($field['value']) && method_exists($field['value'], '__toString')) {
        $value = $field['value']->__toString();
      }
      else {
        $this->logger->error('Found config field %field in %plugin with non-supported value.', [
          '%field' => $field['label'],
          '%plugin' => $plugin->getPluginId(),
        ]);
        $value = '';
      }
      $property = [
        'label' => $field['label'],
        'type' => $field['type'],
        'value' => $value,
        'editable' => $field['editable'] ?? TRUE,
        'binding' => [
          'type' => 'camunda:field',
          'name' => $field['name'],
        ],
      ];
      if (!empty($field['required'])) {
        $property['constraints']['notEmpty'] = TRUE;
      }
      if (isset($field['description'])) {
        $property['description'] = (string) $field['description'];
      }
      if (isset($field['extras'])) {
        /* @noinspection SlowArrayOperationsInLoopInspection */
        $property = array_merge_recursive($property, $field['extras']);
      }
      $properties[] = $property;
    }
    $extraDescriptions = array_unique($extraDescriptions);
    $pluginDefinition = $plugin->getPluginDefinition();
    $template = [
      'name' => (string) $pluginDefinition['label'],
      'id' => 'org.drupal.' . $pluginType . '.' . $plugin->getPluginId(),
      'category' => [
        'id' => $pluginDefinition['provider'],
        'name' => EcaPluginBase::$modules[$pluginDefinition['provider']],
      ],
      'appliesTo' => [$appliesTo],
      'properties' => $properties,
    ];
    if (isset($pluginDefinition['description']) || $extraDescriptions) {
      $template['description'] = ($pluginDefinition['description'] ?? '') . ' ' . implode(' ', $extraDescriptions);
    }
    if ($docUrl) {
      $template['documentationRef'] = $docUrl;
    }
    return $template;
  }

  /**
   * Helper function preparing config fields for events, conditions and actions.
   *
   * @param array $form
   *   The array to which the fields should be added.
   * @param array $extraDescriptions
   *   An array receiving all markup "fields" which can be displayed separately
   *   in the UI.
   *
   * @return array
   *   The prepared config fields.
   */
  protected function prepareConfigFields(array $form, array &$extraDescriptions): array {
    // @todo Add support for nested form fields like e.g. in container/fieldset.
    $fields = [];
    foreach ($form as $key => $definition) {
      if (!is_array($definition)) {
        continue;
      }
      $label = $definition['#title'] ?? $this->convertKeyToLabel($key);
      $description = $definition['#description'] ?? NULL;
      $value = $definition['#default_value'] ?? '';
      $weight = $definition['#weight'] ?? 0;
      $type = 'String';
      $required = $definition['#required'] ?? FALSE;
      // @todo Map to more proper property types of bpmn-js.
      switch ($definition['#type'] ?? 'markup') {

        case 'hidden':
        case 'actions':
          // The modellers can't handle these types, so we ignore them for
          // the templates.
          continue 2;

        case 'item':
        case 'markup':
        case 'container':
          if (isset($definition['#markup'])) {
            $extraDescriptions[] = (string) $definition['#markup'];
          }
          continue 2;

        case 'textarea':
          $type = 'Text';
          break;

        case 'checkbox':
          $fields[] = $this->checkbox($key, $label, $weight, $description, $value);
          continue 2;

        case 'checkboxes':
        case 'radios':
        case 'select':
          if (!is_array($value)) {
            $options = $this->normalizeOptions(form_select_options($definition));
            $fields[] = $this->optionsField($key, $label, $weight, $description, $options, (string) $value, $required);
            continue 2;
          }
          break;

      }
      if (is_bool($value)) {
        $fields[] = $this->checkbox($key, $label, $weight, $description, $value);
        continue;
      }
      if (is_array($value)) {
        $value = implode(',', $value);
      }
      $field = [
        'name' => $key,
        'label' => $label,
        'weight' => $weight,
        'type' => $type,
        'value' => $value,
        'required' => $required,
      ];
      if ($description !== NULL) {
        $field['description'] = $description;
      }
      $fields[] = $field;
    }

    // Sort fields by weight.
    usort($fields, static function ($f1, $f2) {
      $l1 = (int) $f1['weight'];
      $l2 = (int) $f2['weight'];
      if ($l1 < $l2) {
        return -1;
      }
      if ($l1 > $l2) {
        return 1;
      }
      return 0;
    });

    return $fields;
  }

  /**
   * Normalizes an option list into a flat list of keys and labels.
   *
   * This can be called recursively, e.g. for nested option groups.
   *
   * @param array $formApiOptions
   *   The list of options.
   * @param string $prefix
   *   An optional prefix which will be prepended to the label.
   *
   * @return array
   *   The flat option list.
   */
  protected function normalizeOptions(array $formApiOptions, string $prefix = ''): array {
    $options = [];
    foreach ($formApiOptions as $formApiOption) {
      switch ($formApiOption['type']) {
        case 'option':
          $options[$formApiOption['value']] = $prefix . $formApiOption['label'];
          break;

        case 'optgroup':
          $options += $this->normalizeOptions($formApiOption['options'], $formApiOption['label'] . ': ');
          break;

      }
    }
    return $options;
  }

  /**
   * Prepares a field with options as a drop-down.
   *
   * @param string $name
   *   The field name.
   * @param string $label
   *   The field label.
   * @param int $weight
   *   The field weight for sorting.
   * @param string|null $description
   *   The optional field description.
   * @param array $options
   *   Key/value list of available options.
   * @param string $value
   *   The default value for the field.
   * @param bool $required
   *   The setting, if this field is required to be filled by the user.
   *
   * @return array
   *   Prepared option field.
   */
  protected function optionsField(string $name, string $label, int $weight, ?string $description, array $options, string $value, bool $required = FALSE): array {
    $choices = [];
    foreach ($options as $optionValue => $optionName) {
      $choices[] = [
        'name' => (string) $optionName,
        'value' => (string) $optionValue,
      ];
      if ($required && $value === '') {
        $value = (string) $optionValue;
      }
    }
    $field = [
      'name' => $name,
      'label' => $label,
      'weight' => $weight,
      'type' => 'Dropdown',
      'value' => $value,
      'required' => $required,
      'extras' => [
        'choices' => $choices,
      ],
    ];
    if ($description !== NULL) {
      $field['description'] = $description;
    }
    return $field;
  }

  /**
   * Prepares a field as a checkbox.
   *
   * @param string $name
   *   The field name.
   * @param string $label
   *   The field label.
   * @param int $weight
   *   The field weight for sorting.
   * @param string|null $description
   *   The optional field description.
   * @param bool $value
   *   The default value for the field.
   *
   * @return array
   *   Prepared checkbox field.
   */
  protected function checkbox(string $name, string $label, int $weight, ?string $description, bool $value): array {
    $field = [
      'name' => $name,
      'label' => $label,
      'weight' => $weight,
      'type' => 'Dropdown',
      'value' => $value ? 'yes' : 'no',
      'extras' => [
        'choices' => [
          [
            'name' => 'no',
            'value' => 'no',
          ],
          [
            'name' => 'yes',
            'value' => 'yes',
          ],
        ],
      ],
    ];
    if ($description !== NULL) {
      $field['description'] = $description;
    }
    return $field;
  }

  /**
   * Builds a field label from the key.
   *
   * @param string $key
   *   The key of the field from which to build a label.
   *
   * @return string
   *   The built label for the field identified by key.
   */
  protected function convertKeyToLabel(string $key): string {
    $labelParts = explode('_', $key);
    $labelParts[0] = ucfirst($labelParts[0]);
    return implode(' ', $labelParts);
  }

}
